package com.stripe.android.auth

import android.app.Activity
import android.content.Context
import android.os.Parcel
import androidx.test.core.app.ApplicationProvider
import com.google.common.truth.Truth.assertThat
import com.stripe.android.ApiKeyFixtures
import com.stripe.android.PaymentConfiguration
import com.stripe.android.payments.DefaultReturnUrl
import com.stripe.android.payments.StripeBrowserLauncherActivity
import com.stripe.android.view.ActivityScenarioFactory
import com.stripe.android.view.PaymentAuthWebViewActivity
import org.junit.runner.RunWith
import org.robolectric.RobolectricTestRunner
import kotlin.test.BeforeTest
import kotlin.test.Test

@RunWith(RobolectricTestRunner::class)
class PaymentBrowserAuthContractTest {

    private val defaultReturnUrl = DefaultReturnUrl(
        "com.stripe.android.test"
    )

    private val context = ApplicationProvider.getApplicationContext<Context>()
    private val activityScenarioFactory = ActivityScenarioFactory(context)
    private lateinit var activity: Activity

    @BeforeTest
    fun setup() {
        PaymentConfiguration.init(context, ApiKeyFixtures.FAKE_PUBLISHABLE_KEY)

        activityScenarioFactory.createAddPaymentMethodActivity().use { activityScenario ->
            activityScenario.onActivity {
                activity = it
            }
        }
    }

    @Test
    fun `createIntent() when has compatible browser and custom return_url should use PaymentAuthWebViewActivity`() {
        val intent = PaymentBrowserAuthContract().createIntent(
            activity,
            ARGS.copy(
                returnUrl = "myapp://custom"
            )
        )

        assertThat(intent.component?.className)
            .isEqualTo(PaymentAuthWebViewActivity::class.java.name)
    }

    @Test
    fun `createIntent() when has compatible browser and default return_url should use StripeBrowserLauncherActivity`() {
        val intent = PaymentBrowserAuthContract().createIntent(
            activity,
            ARGS.copy(
                returnUrl = defaultReturnUrl.value
            )
        )

        assertThat(intent.component?.className)
            .isEqualTo(StripeBrowserLauncherActivity::class.java.name)
    }

    @Test
    fun `createIntent() when no compatible browser and default return_url should use StripeBrowserLauncherActivity`() {
        val intent = PaymentBrowserAuthContract().createIntent(
            activity,
            ARGS.copy(
                returnUrl = defaultReturnUrl.value
            )
        )

        assertThat(intent.component?.className)
            .isEqualTo(StripeBrowserLauncherActivity::class.java.name)
    }

    @Test
    fun `createIntent() when isInstantApp should use StripeBrowserLauncherActivity`() {
        val intent = PaymentBrowserAuthContract().createIntent(
            activity,
            ARGS.copy(
                isInstantApp = true
            )
        )

        assertThat(intent.component?.className)
            .isEqualTo(StripeBrowserLauncherActivity::class.java.name)
    }

    @Test
    fun `createIntent() should set statusBarColor from activity`() {
        val intent = PaymentBrowserAuthContract().createIntent(
            activity,
            ARGS
        )

        val args = requireNotNull(
            intent.getParcelableExtra<PaymentBrowserAuthContract.Args>("extra_args")
        )
        assertThat(args.statusBarColor)
            .isNotNull()
        assertThat(args.statusBarColor)
            .isEqualTo(activity.window.statusBarColor)
    }

    @Test
    fun `unparcel when no parameters as when started from StripeBrowserLauncherActivity`() {
        val parcel = Parcel.obtain()

        // An NullPointerException is thrown if a constructor doesn't exist that
        // takes a parcel and created the object with empty strings if they
        // don't exist.
        PaymentBrowserAuthContract.Args(parcel)
    }

    private companion object {
        private val ARGS = PaymentBrowserAuthContract.Args(
            clientSecret = "client_secret",
            objectId = "pi_12345",
            requestCode = 5000,
            url = "https://mybank.com/auth",
            returnUrl = "myapp://custom",
            publishableKey = ApiKeyFixtures.FAKE_PUBLISHABLE_KEY,
            isInstantApp = false
        )
    }
}
